@@ -20,6 +20,7 @@
 *     [7] Quinlan, J. Ross. C4. 5: programs for machine learning. Elsevier, 2014.
 *     [8] Cortes, David. "Distance approximation using Isolation Forests." arXiv preprint arXiv:1910.12362 (2019).
 *     [9] Cortes, David. "Imputing missing values with unsupervised random trees." arXiv preprint arXiv:1911.06646 (2019).
+*     [10] Cortes, David. "Revisiting randomized choices in isolation forests." arXiv preprint arXiv:2110.13402 (2021).
 * 
 *     BSD 2-Clause License
 *     Copyright (c) 2019-2021, David Cortes
@@ -354,9 +355,7 @@ void split_hplane_recursive(std::vector<IsoHPlane>   &hplanes,
         if (workspace.xmin == workspace.xmax)
             goto terminal_statistics; /* in theory, could try again too, this could just be an unlucky case */
         
-        hplanes.back().split_point =
-            std::uniform_real_distribution<double>(workspace.xmin, workspace.xmax)
-                                                  (workspace.rnd_generator);
+        hplanes.back().split_point = sample_random_uniform(workspace.xmin, workspace.xmax, workspace.rnd_generator);
 
         /* determine acceptable range */
         if (model_params.penalize_range)
