@@ -20,6 +20,7 @@
 *     [7] Quinlan, J. Ross. C4. 5: programs for machine learning. Elsevier, 2014.
 *     [8] Cortes, David. "Distance approximation using Isolation Forests." arXiv preprint arXiv:1910.12362 (2019).
 *     [9] Cortes, David. "Imputing missing values with unsupervised random trees." arXiv preprint arXiv:1911.06646 (2019).
+*     [10] Cortes, David. "Revisiting randomized choices in isolation forests." arXiv preprint arXiv:2110.13402 (2021).
 * 
 *     BSD 2-Clause License
 *     Copyright (c) 2019-2021, David Cortes
@@ -99,18 +100,58 @@ void merge_models(IsoForest*     model,      IsoForest*     other,
                   ExtIsoForest*  ext_model,  ExtIsoForest*  ext_other,
                   Imputer*       imputer,    Imputer*       iother)
 {
-    if (model != NULL && other != NULL)
-        model->trees.insert(model->trees.end(),
-                            other->trees.begin(),
-                            other->trees.end());
+    size_t curr_size_model = (model != NULL)? (model->trees.size()) : 0;
+    size_t curr_size_model_ext = (ext_model != NULL)? (ext_model->hplanes.size()) : 0;
+    size_t curr_size_imputer = (imputer != NULL)? (imputer->imputer_tree.size()) : 0;
 
-    if (ext_model != NULL && ext_other != NULL)
-        ext_model->hplanes.insert(ext_model->hplanes.end(),
-                                  ext_other->hplanes.begin(),
-                                  ext_other->hplanes.end());
+    try
+    {
+        if (model != NULL && other != NULL)
+        {
+            if (model == other)
+            {
+                auto other_copy = *other;
+                merge_models(model, &other_copy, ext_model, ext_other, imputer, iother);
+                return;
+            }
+            model->trees.insert(model->trees.end(),
+                                other->trees.begin(),
+                                other->trees.end());
 
-    if (imputer != NULL && iother != NULL)
-        imputer->imputer_tree.insert(imputer->imputer_tree.end(),
-                                     iother->imputer_tree.begin(),
-                                     iother->imputer_tree.end());
+        }
+
+        if (ext_model != NULL && ext_other != NULL)
+        {
+            if (ext_model == ext_other)
+            {
+                auto other_copy = *ext_other;
+                merge_models(model, other, ext_model, &other_copy, imputer, iother);
+                return;
+            }
+            ext_model->hplanes.insert(ext_model->hplanes.end(),
+                                      ext_other->hplanes.begin(),
+                                      ext_other->hplanes.end());
+        }
+
+        if (imputer != NULL && iother != NULL)
+        {
+            if (imputer == iother)
+            {
+                auto other_copy = *iother;
+                merge_models(model, other, ext_model, ext_other, imputer, &other_copy);
+                return;
+            }
+            imputer->imputer_tree.insert(imputer->imputer_tree.end(),
+                                         iother->imputer_tree.begin(),
+                                         iother->imputer_tree.end());
+        }
+    }
+
+    catch(...)
+    {
+        if (model != NULL) model->trees.resize(curr_size_model);
+        if (ext_model != NULL) ext_model->hplanes.resize(curr_size_model_ext);
+        if (imputer != NULL) imputer->imputer_tree.resize(curr_size_imputer);
+        throw;
+    }
 }
