@@ -194,7 +194,8 @@ void predict_iforest(real_t numeric_data[], int categ_data[],
         if (
             model_outputs->missing_action == Fail &&
             (model_outputs->new_cat_action != Weighted || prediction_data.categ_data == NULL) &&
-            prediction_data.Xc_indptr == NULL && prediction_data.Xr_indptr == NULL
+            prediction_data.Xc_indptr == NULL && prediction_data.Xr_indptr == NULL &&
+            !model_outputs->has_range_penalty
             )
         {
             #pragma omp parallel for if(nrows > 1) schedule(static) num_threads(nthreads) \
@@ -246,7 +247,8 @@ void predict_iforest(real_t numeric_data[], int categ_data[],
             model_outputs_ext->missing_action == Fail &&
             prediction_data.categ_data == NULL &&
             prediction_data.Xc_indptr == NULL &&
-            prediction_data.Xr_indptr == NULL
+            prediction_data.Xr_indptr == NULL &&
+            !model_outputs_ext->has_range_penalty
             )
         {
             #pragma omp parallel for if(nrows > 1) schedule(static) num_threads(nthreads) \
@@ -360,7 +362,6 @@ void traverse_itree_no_recurse(std::vector<IsoTree>  &tree,
                                     :
                                 (tree[curr_lev].col_num + row * prediction_data.ncols_numeric)
                             ];
-                    output_depth -= (xval < tree[curr_lev].range_low) || (xval > tree[curr_lev].range_high);
                     curr_lev = (xval <= tree[curr_lev].num_split)?
                                 tree[curr_lev].tree_left : tree[curr_lev].tree_right;
                     break;
@@ -400,6 +401,8 @@ void traverse_itree_no_recurse(std::vector<IsoTree>  &tree,
                                 {
                                     case Random:
                                     {
+                                        cval = (cval >= (int)tree[curr_lev].cat_split.size())?
+                                                (cval % (int)tree[curr_lev].cat_split.size()) : cval;
                                         curr_lev = (tree[curr_lev].cat_split[cval])?
                                                     tree[curr_lev].tree_left : tree[curr_lev].tree_right;
                                         break;
@@ -580,7 +583,7 @@ double traverse_itree(std::vector<IsoTree>     &tree,
                                 prediction_data.is_col_major?
                                 (row +  tree[curr_lev].col_num * prediction_data.nrows)
                                     :
-                                (tree[curr_lev].col_num * row * prediction_data.ncols_categ)
+                                (tree[curr_lev].col_num + row * prediction_data.ncols_categ)
                             ];
                     if (cval < 0)
                     {
@@ -675,6 +678,8 @@ double traverse_itree(std::vector<IsoTree>     &tree,
                                     {
                                         case Random:
                                         {
+                                            cval = (cval >= (int)tree[curr_lev].cat_split.size())?
+                                                    (cval % (int)tree[curr_lev].cat_split.size()) : cval;
                                             curr_lev = (tree[curr_lev].cat_split[cval])?
                                                         tree[curr_lev].tree_left : tree[curr_lev].tree_right;
                                             break;
@@ -740,7 +745,7 @@ double traverse_itree(std::vector<IsoTree>     &tree,
 }
 
 /* this is a simpler version for situations in which there is
-   only numeric data in dense arrays and no missing values */
+   only numeric data in dense arrays, no missing values, no range penalty */
 template <class PredictionData, class sparse_ix>
 [[gnu::hot]]
 void traverse_hplane_fast(std::vector<IsoHPlane>  &hplane,
@@ -784,8 +789,6 @@ void traverse_hplane_fast(std::vector<IsoHPlane>  &hplane,
             }
         }
 
-        output_depth -= (hval < hplane[curr_lev].range_low) ||
-                        (hval > hplane[curr_lev].range_high);
         curr_lev      = (hval <= hplane[curr_lev].split_point)?
                          hplane[curr_lev].hplane_left : hplane[curr_lev].hplane_right;
     }
@@ -939,9 +942,22 @@ void traverse_hplane(std::vector<IsoHPlane>   &hplane,
                                 case SubSet:
                                 {
                                     if (cval >= (int)hplane[curr_lev].cat_coef[ncols_categ].size())
-                                        hval += hplane[curr_lev].fill_new[ncols_categ];
+                                    {
+                                        if (model_outputs.new_cat_action == Random) {
+                                            cval = cval % (int)hplane[curr_lev].cat_coef[ncols_categ].size();
+                                            hval += hplane[curr_lev].cat_coef[ncols_categ][cval];
+                                        }
+
+                                        else {
+                                            hval += hplane[curr_lev].fill_new[ncols_categ];
+                                        }
+                                    }
+                                    
                                     else
+                                    {
                                         hval += hplane[curr_lev].cat_coef[ncols_categ][cval];
+                                    }
+                                    
                                     break;
                                 }
                             }
@@ -982,29 +998,12 @@ void batched_csc_predict(PredictionData<real_t, sparse_ix> &prediction_data, int
     nthreads = 1;
     #endif
     std::vector<WorkerForPredictCSC> worker_memory(nthreads);
-    bool has_range_penalty = true;
 
     bool threw_exception = false;
     std::exception_ptr ex = NULL;
 
     if (model_outputs != NULL)
     {
-        for (auto &tree : model_outputs->trees)
-        {
-            for (auto &node : tree)
-            {
-                if (node.score < 0 && node.col_type == Numeric)
-                {
-                    if (isinf(node.range_low) && isinf(node.range_high))
-                        has_range_penalty = false;
-                    goto got_range_penalty;
-                }
-            }
-
-            got_range_penalty:
-            {}
-        }
-
         #pragma omp parallel for schedule(dynamic) num_threads(nthreads) \
                 shared(worker_memory, model_outputs, prediction_data, tree_num, per_tree_depths, threw_exception)
         for (size_t_for tree = 0; tree < (decltype(tree))model_outputs->trees.size(); tree++)
@@ -1021,8 +1020,11 @@ void batched_csc_predict(PredictionData<real_t, sparse_ix> &prediction_data, int
                               ptr_worker->ix_arr.end(),
                               (size_t)0);
 
-                    if (model_outputs->missing_action == Divide)
+                    if (model_outputs->missing_action == Divide ||
+                        (model_outputs->new_cat_action == Weighted && prediction_data.categ_data != NULL)
+                    ) {
                         ptr_worker->weights_arr.resize(prediction_data.nrows);
+                    }
                 }
 
                 ptr_worker->st  = 0;
@@ -1040,7 +1042,7 @@ void batched_csc_predict(PredictionData<real_t, sparse_ix> &prediction_data, int
                                         ((sparse_ix*)NULL) : (tree_num + tree*prediction_data.nrows),
                                    per_tree_depths,
                                    (size_t)0,
-                                   has_range_penalty);
+                                   model_outputs->has_range_penalty);
             }
 
             catch(...)
@@ -1059,22 +1061,6 @@ void batched_csc_predict(PredictionData<real_t, sparse_ix> &prediction_data, int
 
     else
     {
-        for (auto &tree : model_outputs_ext->hplanes)
-        {
-            for (auto &node : tree)
-            {
-                if (node.score < 0)
-                {
-                    if (isinf(node.range_low) && isinf(node.range_high))
-                        has_range_penalty = false;
-                    goto got_range_penalty2;
-                }
-            }
-
-            got_range_penalty2:
-            {}
-        }
-
         #pragma omp parallel for schedule(dynamic) num_threads(nthreads) \
                 shared(worker_memory, model_outputs_ext, prediction_data, tree_num, per_tree_depths, threw_exception)
         for (size_t_for tree = 0; tree < (decltype(tree))model_outputs_ext->hplanes.size(); tree++)
@@ -1104,7 +1090,7 @@ void batched_csc_predict(PredictionData<real_t, sparse_ix> &prediction_data, int
                                         ((sparse_ix*)NULL) : (tree_num + tree*prediction_data.nrows),
                                     per_tree_depths,
                                     (size_t)0,
-                                    has_range_penalty);
+                                    model_outputs_ext->has_range_penalty);
             }
 
             catch(...)
@@ -1235,6 +1221,8 @@ void traverse_itree_csc(WorkerForPredictCSC   &workspace,
     }
 
     /* continue splitting recursively */
+    if (model_outputs.new_cat_action == Weighted)
+        goto missing_action_divide;
     switch(model_outputs.missing_action)
     {
         case Impute:
@@ -1294,6 +1282,7 @@ void traverse_itree_csc(WorkerForPredictCSC   &workspace,
 
         case Divide:
         {
+            missing_action_divide:
             /* TODO: maybe here it shouldn't copy the whole ix_arr,
                but then it'd need to re-generate it from outside too */
             std::vector<double> weights_arr;
